# credits to slimakoi for amino.py base
import base64
from binascii import hexlify
import json
from os import urandom
from time import timezone
from time import time as timestamp
from typing import BinaryIO, Union
from uuid import UUID

from json_minify import json_minify

from .client import Client
from . import entities, exceptions


class SubClient(Client):
    def __init__(self, comId: str = None, aminoId: str = None, *, profile: entities.UserProfile):
        Client.__init__(self)
        self.vc_connect = False

        if comId is not None:
            self.comId = comId
            self.community: entities.Community = self.get_community_info(comId)

        if aminoId is not None:
            self.comId = Client().search_community(aminoId).comId[0]
            self.community: entities.Community = Client().get_community_info(self.comId)

        if comId is None and aminoId is None:
            raise exceptions.NoCommunity()

        try:
            self.profile: entities.UserProfile = self.get_user_info(userId=profile.userId)
        except AttributeError:
            raise exceptions.FailedLogin()
        except exceptions.UserUnavailable:
            pass

    def check_in(self, tz: int = -timezone // 1000):
        url = f"{self.api}/x{self.comId}/s/check-in"
        data = {
            "timezone": tz,
            "timestamp": int(timestamp() * 1000)
        }
        return self._post(url, data)

    def lottery(self, tz: int = -timezone // 1000):
        url = f"{self.api}/x{self.comId}/s/check-in/lottery"
        data = {
            "timezone": tz,
            "timestamp": int(timestamp() * 1000)
        }
        json_object = self._post_get_json(url, data)
        return json_object and entities.LotteryLog(json_object["lotteryLog"]).Self

    def edit_profile(self, nickname: str = None, content: str = None, icon: BinaryIO = None, chatRequestPrivilege: str = None, imageList: list = None, captionList: list = None, backgroundImage: str = None, backgroundColor: str = None, titles: list = None, colors: list = None, defaultBubbleId: str = None):
        mediaList = []

        url = f"{self.api}/x{self.comId}/s/user-profile/{self.profile.userId}"
        data = {"timestamp": int(timestamp() * 1000)}

        if captionList is not None:
            for image, caption in zip(imageList, captionList):
                mediaList.append([100, self.upload_media(image, "image"), caption])

        else:
            if imageList is not None:
                for image in imageList:
                    mediaList.append([100, self.upload_media(image, "image"), None])

        if imageList is not None or captionList is not None:
            data["mediaList"] = mediaList

        if nickname:
            data["nickname"] = nickname
        if icon:
            data["icon"] = self.upload_media(icon, "image")
        if content:
            data["content"] = content

        if chatRequestPrivilege:
            data["extensions"] = {"privilegeOfChatInviteRequest": chatRequestPrivilege}
        if backgroundImage:
            data["extensions"] = {"style": {"backgroundMediaList": [[100, backgroundImage, None, None, None]]}}
        if backgroundColor:
            data["extensions"] = {"style": {"backgroundColor": backgroundColor}}
        if defaultBubbleId:
            data["extensions"] = {"defaultBubbleId": defaultBubbleId}

        if titles or colors:
            tlt = []
            for titles, colors in zip(titles, colors):
                tlt.append({"title": titles, "color": colors})

            data["extensions"] = {"customTitles": tlt}

        return self._post(url, data)

    def like_blog(self, blogId: Union[str, list] = None, wikiId: str = None):
        """
        Like a Blog, Multiple Blogs or a Wiki.

        **Parameters**
            - **blogId** : ID of the Blog or List of IDs of the Blogs. (for Blogs)
            - **wikiId** : ID of the Wiki. (for Wikis)

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        data = {
            "value": 4,
            "timestamp": int(timestamp() * 1000)
        }

        if blogId:
            if isinstance(blogId, str):
                url = f"{self.api}/x{self.comId}/s/blog/{blogId}/vote?cv=1.2"
                data["eventSource"] = "UserProfileView"
            elif isinstance(blogId, list):
                url = f"{self.api}/x{self.comId}/s/feed/vote"
                data["targetIdList"] = blogId
            else:
                raise exceptions.WrongType
        elif wikiId:
            url = f"{self.api}/x{self. comId}/s/item/{wikiId}/vote?cv=1.2"
            data["eventSource"] = "PostDetailView"
        else:
            raise exceptions.SpecifyType()

        return self._post(url, data)

    def send_active_obj(self, startTime: int = None, endTime: int = None, optInAdsFlags: int = 2147483647, tz: int = -timezone // 1000, timers: list = None, timestamp: int = int(timestamp() * 1000)):
        url = f"{self.api}/x{self.comId}/s/community/stats/user-active-time"
        data = {"userActiveTimeChunkList": [{"start": startTime, "end": endTime}], "timestamp": timestamp, "optInAdsFlags": optInAdsFlags, "timezone": tz}
        if timers:
            data["userActiveTimeChunkList"] = timers
        data = json_minify(json.dumps(data))
        with TorRequests() as client:
            with client.get_session() as session:
                headers = self.parse_headers(data=data)
                # NOTE: changed requests.post to session.post. Is this right?
                response = session.post(url, headers=headers, data=data, proxies=self.proxies, verify=self.certificatePath)
                if response.status_code != 200:
                    return exceptions.CheckException(json.loads(response.text))

                return response.status_code

    def activity_status(self, status: str):
        if "on" in status.lower():
            status = 1
        elif "off" in status.lower():
            status = 2
        else:
            raise exceptions.WrongType(status)

        url = f"{self.api}/x{self.comId}/s/user-profile/{self.profile.userId}/online-status"
        data = {
            "onlineStatus": status,
            "duration": 86400,
            "timestamp": int(timestamp() * 1000)
        }
        return self._post(url, data)

    def watch_ad(self):
        url = f"{self.api}/g/s/wallet/ads/video/start"
        return self._post(url)

    def start_chat(self, userId: Union[str, list], message: str, title: str = None, content: str = None, isGlobal: bool = False, publishToGlobal: bool = False):
        if isinstance(userId, str):
            userIds = [userId]
        elif isinstance(userId, list):
            userIds = userId
        else:
            raise exceptions.WrongType(type(userId))

        url = f"{self.api}/x{self.comId}/s/chat/thread"
        data = {
            "title": title,
            "inviteeUids": userIds,
            "initialMessageContent": message,
            "content": content,
            "timestamp": int(timestamp() * 1000)
        }

        if isGlobal:
            data["type"] = 2
            data["eventSource"] = "GlobalComposeMenu"
        else:
            data["type"] = 0

        data["publishToGlobal"] = int(publishToGlobal)
        return self._post(url, data)

    def invite_to_chat(self, userId: Union[str, list], chatId: str):
        if isinstance(userId, str):
            userIds = [userId]
        elif isinstance(userId, list):
            userIds = userId
        else:
            raise exceptions.WrongType(type(userId))

        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/member/invite"
        data = {
            "uids": userIds,
            "timestamp": int(timestamp() * 1000)
        }
        return self._post(url, data)

    def send_coins(self, coins: int, blogId: str = None, chatId: str = None, objectId: str = None, transactionId: str = None):
        url = None
        if transactionId is None:
            transactionId = str(UUID(hexlify(urandom(16)).decode('ascii')))

        data = {
            "coins": coins,
            "tippingContext": {"transactionId": transactionId},
            "timestamp": int(timestamp() * 1000)
        }

        if blogId is not None:
            url = f"{self.api}/x{self.comId}/s/blog/{blogId}/tipping"
        if chatId is not None:
            url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/tipping"
        if objectId is not None:
            data["objectId"] = objectId
            data["objectType"] = 2
            url = f"{self.api}/x{self.comId}/s/tipping"

        if url is None:
            raise exceptions.SpecifyType()

        return self._post(url, data)

    def thank_tip(self, chatId: str, userId: str):
        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/tipping/tipped-users/{userId}/thank"
        return self._post(url)

    def follow(self, userId: Union[str, list]):
        """
        Follow an User or Multiple Users.

        **Parameters**
            - **userId** : ID of the User or List of IDs of the Users.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        if isinstance(userId, str):
            url = f"{self.api}/x{self.comId}/s/user-profile/{userId}/member"
            return None if self._post(url) is None else 200
        elif isinstance(userId, list):
            url = f"{self.api}/x{self.comId}/s/user-profile/{self.profile.userId}/joined"
            data = json.dumps({"targetUidList": userId, "timestamp": int(timestamp() * 1000)})
            return None if self._post(url, data) is None else 200
        else:
            raise exceptions.WrongType(type(userId))

    def unfollow(self, userId: str):
        """
        Unfollow an User.

        **Parameters**
            - **userId** : ID of the User.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/user-profile/{self.profile.userId}/joined/{userId}"
        return self._delete(url)

    def send_message(self, chatId: str, message: str = None, messageType: int = 0, file: BinaryIO = None, fileType: str = None, replyTo: str = None, mentionUserIds: list = None, stickerId: str = None, embedId: str = None, embedType: int = None, embedLink: str = None, embedTitle: str = None, embedContent: str = None, embedImage: BinaryIO = None):
        """
        Send a Message to a Chat.

        **Parameters**
            - **message** : Message to be sent
            - **chatId** : ID of the Chat.
            - **file** : File to be sent.
            - **fileType** : Type of the file.
                - ``audio``, ``image``, ``gif``
            - **messageType** : Type of the Message.
            - **mentionUserIds** : List of User IDS to mention. '@' needed in the Message.
            - **replyTo** : Message ID to reply to.
            - **stickerId** : Sticker ID to be sent.
            - **embedTitle** : Title of the Embed.
            - **embedContent** : Content of the Embed.
            - **embedLink** : Link of the Embed.
            - **embedImage** : Image of the Embed.
            - **embedId** : ID of the Embed.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """

        if message is not None and file is None:
            message = message.replace("<$", "‎‏").replace("$>", "‬‭")

        mentions = []
        if mentionUserIds:
            for mention_uid in mentionUserIds:
                mentions.append({"uid": mention_uid})

        if embedImage:
            embedImage = [[100, self.upload_media(embedImage, "image"), None]]

        data = {
            "type": messageType,
            "content": message,
            "clientRefId": int(timestamp() / 10 % 1000000000),
            "attachedObject": {
                "objectId": embedId,
                "objectType": embedType,
                "link": embedLink,
                "title": embedTitle,
                "content": embedContent,
                "mediaList": embedImage
            },
            "extensions": {"mentionedArray": mentions},
            "timestamp": int(timestamp() * 1000)
        }

        if replyTo: data["replyMessageId"] = replyTo

        if stickerId:
            data["content"] = None
            data["stickerId"] = stickerId
            data["type"] = 3

        if file:
            data["content"] = None
            if fileType == "audio":
                data["type"] = 2
                data["mediaType"] = 110

            elif fileType == "image":
                data["mediaType"] = 100
                data["mediaUploadValueContentType"] = "image/jpg"
                data["mediaUhqEnabled"] = True

            elif fileType == "gif":
                data["mediaType"] = 100
                data["mediaUploadValueContentType"] = "image/gif"
                data["mediaUhqEnabled"] = True

            else: raise exceptions.SpecifyType(fileType)

            data["mediaUploadValue"] = base64.b64encode(file.read()).decode()

        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/message"
        return self._post(url)

    def delete_message(self, chatId: str, messageId: str, asStaff: bool = False, reason: str = None):
        """
        Delete a Message from a Chat.

        **Parameters**
            - **messageId** : ID of the Message.
            - **chatId** : ID of the Chat.
            - **asStaff** : If execute as a Staff member (Leader or Curator).
            - **reason** : Reason of the action to show on the Moderation History.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        if not asStaff:
            url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/message/{messageId}"
            return self._delete(url)

        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/message/{messageId}/admin"
        data = {
            "adminOpName": 102,
            "timestamp": int(timestamp() * 1000)
        }
        if reason:
            data["adminOpNote"] = {"content": reason}

        return self._post(url, data)

    def join_chat(self, chatId: str):
        """
        Join an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/member/{self.profile.userId}"
        return self._post(url)

    def leave_chat(self, chatId: str):
        """
        Leave an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.

        **Returns**
            - **Success** : 200 (int)

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/member/{self.profile.userId}"
        return self._delete(url)

    def subscribe(self, userId: str, autoRenew: str = False, transactionId: str = None):
        if transactionId is None:
            transactionId = str(UUID(hexlify(urandom(16)).decode('ascii')))

        url = f"{self.api}/x{self.comId}/s/influencer/{userId}/subscribe"
        data = {
            "paymentContext": {
                "transactionId": transactionId,
                "isAutoRenew": autoRenew
            },
            "timestamp": int(timestamp() * 1000)
        }
        return self._post(url, data)

    def get_vc_reputation_info(self, chatId: str):
        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/avchat-reputation"
        return self._get_entity(url, entities.VcReputation)

    def claim_vc_reputation(self, chatId: str):
        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/avchat-reputation"
        return self._get_entity(url, entities.VcReputation)

    def get_all_users(self, type: str = "recent", start: int = 0, size: int = 25):
        if type == "recent":
            url = f"{self.api}/x{self.comId}/s/user-profile?type=recent&start={start}&size={size}"
        elif type == "banned":
            url = f"{self.api}/x{self.comId}/s/user-profile?type=banned&start={start}&size={size}"
        elif type == "featured":
            url = f"{self.api}/x{self.comId}/s/user-profile?type=featured&start={start}&size={size}"
        elif type == "leaders":
            url = f"{self.api}/x{self.comId}/s/user-profile?type=leaders&start={start}&size={size}"
        elif type == "curators":
            url = f"{self.api}/x{self.comId}/s/user-profile?type=curators&start={start}&size={size}"
        else:
            raise exceptions.WrongType(type)

        return self._get_entity(url, entities.UserProfileList)

    def get_online_users(self, start: int = 0, size: int = 25):
        url = f"{self.api}/x{self.comId}/s/live-layer?topic=ndtopic:x{self.comId}:online-members&start={start}&size={size}"
        return self._get_entity(url, entities.UserProfileCountList)

    def get_online_favorite_users(self, start: int = 0, size: int = 25):
        url = f"{self.api}/x{self.comId}/s/user-group/quick-access?type=online&start={start}&size={size}"
        return self._get_entity(url, entities.UserProfileCountList)

    def get_user_info(self, userId: str):
        """
        Information of an User.

        **Parameters**
            - **userId** : ID of the User.

        **Returns**
            - **Success** : :meth:`User Object <amino.lib.util.entities.UserProfile>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/user-profile/{userId}"
        return self._get_entity(url, entities.UserProfile)

    def get_user_following(self, userId: str, start: int = 0, size: int = 25):
        """
        List of Users that the User is Following.

        **Parameters**
            - **userId** : ID of the User.
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`User List <amino.lib.util.entities.UserProfileList>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/user-profile/{userId}/joined?start={start}&size={size}"
        return self._get_entity(url, entities.UserProfileList, "userProfileList")

    def get_user_followers(self, userId: str, start: int = 0, size: int = 25):
        """
        List of Users that are Following the User.

        **Parameters**
            - **userId** : ID of the User.
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`User List <amino.lib.util.entities.UserProfileList>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/user-profile/{userId}/member?start={start}&size={size}"
        return self._get_entity(url, entities.UserProfileList, "userProfileList")

    def get_user_checkins(self, userId: str):
        url = f"{self.api}/x{self.comId}/s/check-in/stats/{userId}?timezone={-timezone // 1000}"
        return self._get_entity(url, entities.UserCheckIns)

    def get_user_blogs(self, userId: str, start: int = 0, size: int = 25):
        url = f"{self.api}/x{self.comId}/s/blog?type=user&q={userId}&start={start}&size={size}"
        return self._get_entity(url, entities.BlogList, "blogList")

    def get_user_achievements(self, userId: str):
        url = f"{self.api}/x{self.comId}/s/user-profile/{userId}/achievements"
        return self._get_entity(url, entities.UserAchievements, "achievements")

    def get_influencer_fans(self, userId: str, start: int = 0, size: int = 25):
        url = f"{self.api}/x{self.comId}/s/influencer/{userId}/fans?start={start}&size={size}"
        return self._get_entity(url, entities.InfluencerFans)

    def search_users(self, nickname: str, start: int = 0, size: int = 25):
        url = f"{self.api}/x{self.comId}/s/user-profile?type=name&q={nickname}&start={start}&size={size}"
        return self._get_entity(url, entities.UserProfileList, "userProfileList")

    def get_tipped_users(self, blogId: str = None, wikiId: str = None, quizId: str = None, fileId: str = None, chatId: str = None, start: int = 0, size: int = 25):
        if blogId or quizId:
            if quizId is not None:
                blogId = quizId

            url = f"{self.api}/x{self.comId}/s/blog/{blogId}/tipping/tipped-users-summary?start={start}&size={size}"
        elif wikiId:
            url = f"{self.api}/x{self.comId}/s/item/{wikiId}/tipping/tipped-users-summary?start={start}&size={size}"
        elif chatId:
            url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/tipping/tipped-users-summary?start={start}&size={size}"
        elif fileId:
            url = f"{self.api}/x{self.comId}/s/shared-folder/files/{fileId}/tipping/tipped-users-summary?start={start}&size={size}"
        else:
            raise exceptions.SpecifyType()

        return self._get_entity(url, entities.TippedUsersSummary)

    def get_chat_threads(self, start: int = 0, size: int = 25):
        """
        List of Chats the account is in.

        **Parameters**
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`Chat List <amino.lib.util.entities.ThreadList>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/chat/thread?type=joined-me&start={start}&size={size}"
        return self._get_entity(url, entities.ThreadList, "threadList")

    def get_public_chat_threads(self, type: str = "recommended", start: int = 0, size: int = 25):
        """
        List of Public Chats of the Community.

        **Parameters**
            - *start* : Where to start the list.
            - *size* : Size of the list.

        **Returns**
            - **Success** : :meth:`Chat List <amino.lib.util.entities.ThreadList>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/chat/thread?type=public-all&filterType={type}&start={start}&size={size}"
        return self._get_entity(url, entities.ThreadList, "threadList")

    def get_chat_thread(self, chatId: str):
        """
        Get the Chat Object from an Chat ID.

        **Parameters**
            - **chatId** : ID of the Chat.

        **Returns**
            - **Success** : :meth:`Chat Object <amino.lib.util.entities.Thread>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}"
        return self._get_entity(url, entities.Thread, "thread")

    def get_chat_messages(self, chatId: str, size: int = 25, pageToken: str = None):
        """
        List of Messages from an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.
            - *size* : Size of the list.
            - *pageToken* : Next Page Token.

        **Returns**
            - **Success** : :meth:`Message List <amino.lib.util.entities.MessageList>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        if pageToken is not None:
            url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/message?v=2&pagingType=t&pageToken={pageToken}&size={size}"
        else:
            url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/message?v=2&pagingType=t&size={size}"

        return self._get_entity(url, entities.GetMessages)

    def get_message_info(self, chatId: str, messageId: str):
        """
        Information of an Message from an Chat.

        **Parameters**
            - **chatId** : ID of the Chat.
            - **message** : ID of the Message.

        **Returns**
            - **Success** : :meth:`Message Object <amino.lib.util.entities.Message>`

            - **Fail** : :meth:`Exceptions <amino.lib.util.exceptions>`
        """
        url = f"{self.api}/x{self.comId}/s/chat/thread/{chatId}/message/{messageId}"
        return self._get_entity(url, entities.Message, "message")

    def get_blog_info(self, blogId: str = None, wikiId: str = None, quizId: str = None, fileId: str = None):
        if blogId or quizId:
            if quizId is not None:
                blogId = quizId
                url = f"{self.api}/x{self.comId}/s/blog/{blogId}"
                return self._get_entity(url, entities.GetBlogInfo)
        elif wikiId:
            url = f"{self.api}/x{self.comId}/s/item/{wikiId}"
            return self._get_entity(url, entities.GetWikiInfo)
        elif fileId:
            url = f"{self.api}/x{self.comId}/s/shared-folder/files/{fileId}"
            return self._get_entity(url, entities.SharedFolderFile, "file")
        else:
            raise exceptions.SpecifyType()

    def ban(self, userId: str, reason: str, banType: int = None):
        url = f"{self.api}/x{self.comId}/s/user-profile/{userId}/ban"
        data = {
            "reasonType": banType,
            "note": {
                "content": reason
            },
            "timestamp": int(timestamp() * 1000)
        }
        return self._post_get_json(url, data)

    def unban(self, userId: str, reason: str):
        url = f"{self.api}/x{self.comId}/s/user-profile/{userId}/unban"
        data = {
            "note": {
                "content": reason
            },
            "timestamp": int(timestamp() * 1000)
        }
        return self._post_get_json(url, data)

    def send_action(self, actions: list, blogId: str = None, quizId: str = None, lastAction: bool = False):
        t = 306 if lastAction else 304
        data = {
            "o": {
                "actions": actions,
                "target": f"ndc://x{self.comId}/",
                "ndcId": int(self.comId),
                "params": {"topicIds": [45841, 17254, 26542, 42031, 22542, 16371, 6059, 41542, 15852]},
                "id": "831046"
            },
            "t": t
        }

        if blogId is not None or quizId is not None:
            data["target"] = f"ndc://x{self.comId}/blog/{blogId}"
            if blogId is not None:
                data["params"]["blogType"] = 0
            if quizId is not None:
                data["params"]["blogType"] = 6

        return self.send(json.dumps(data))
